#include	<Input.h>
#include	<InputServerMethod.h>
#include	<Messenger.h>
#include	<String.h>

static void
hexdump(const char *src, int32 srcLen)
{
	for (int32 i= 0; i < srcLen; i++) {
		if ((uchar)src[i] < 16)
			printf("0%x ", src[i]);
		else
			printf("%x ", (uchar)src[i]);
	}
}

static int32
utf8_str_size(const char* src) 
{
	return  ((uchar)src[0] < 0xc0) 					? 1 :
				 (((uchar)src[0] & 0xf0) == 0xe0) ? 3 :
				 (((uchar)src[0] & 0xe0) == 0xc0) ? 2 :
				 (((uchar)src[0] & 0xf8) == 0xf0) ? 4 :
				 (((uchar)src[0] & 0xfc) == 0xf8) ? 5 : 6;
}

static inline uint32 utf8_str_len(const char* ustring, int32 length) 
{
	uint32 cnt = 0;
	for (int32 i= 0; i < length && ustring[i] != '\0'; cnt++) {
		i += utf8_str_size(ustring+i);
	}
	return cnt;       
}

static void
cutf16(const char *src, int32 *srcLen, PRUnichar *dst, int32 *dstLen)
{
	printf("cutf8(%ld %ld: ", *srcLen, *dstLen);
	int32 s = 0, d = 0;
	while (s < *srcLen && d < *dstLen) {
		if ((uchar)src[s] < 0x80)
			dst[d++] = src[s++];
		else {
			int32 len  = utf8_str_size(src+s)-1;
			uint16 u16 = ((uchar)src[s++] & (0x3f >> len));
			for (int32 i = 0; i < len; i++) {
				u16 = (u16 << 6) + ((uchar)src[s++] & 0x3f);
			}
			dst[d++] = u16;
		}
		printf("%x ", dst[d-1]);
	}
	dst[d] = 0;
	*srcLen = s;
	*dstLen = d;
	printf(") %ld %ld\n", *srcLen, *dstLen);
}

void nsViewBeOS::DoIME(BMessage *msg)
{
	nsWindow	*w = (nsWindow *)GetMozillaWidget();
	nsToolkit	*t;

	if(w && (t = w->GetToolkit()) != 0) {
		ssize_t size = msg->FlattenedSize();
		int32		argc = (size+3)/4;
		uint32 *args = new uint32[argc];
		status_t state = msg->Flatten((char*)args, size);
		printf("Flatten(%ld < %ld) %ld\n ",  msg->FlattenedSize(), argc*4, state);
//		hexdump((const char*)(args+15), 16);
//		BMessage chk;
//		state = chk.Unflatten((const char*)args);
//		printf("Chk %lx\n", state);
		MethodInfo *info = new MethodInfo(w, w, nsWindow::IME, argc, args);
		t->CallMethodAsync(info);
		NS_RELEASE(t);
		delete[] args;
	}
}






class BezillaIME
{
	public:
										BezillaIME();
							void	RunIME(uint32 *args, nsWindow *owner, BView* view);
							void	DispatchText(PRUnichar* text, PRUint32 txtCount, nsTextRange* txtRuns);
							void	DispatchIME(PRUint32 what);
							void	DispatchCancelIME();
						PRBool	DispatchWindowEvent(nsGUIEvent* event);

	private:
				nsWindow*		mTarget;
						BView*	mView;
				BMessenger	mIME;
				PRUnichar*	mText;
						int32		mCount;
					BPoint		mCaret;
				PRUint32		mStateIME;

};

	BezillaIME 		be_IME;			// global

BezillaIME::BezillaIME()
	: mTarget(NULL)
	, mText(NULL), mCount(0)
	, mStateIME(NS_COMPOSITION_END)
{
}
								
void
BezillaIME::RunIME(uint32 *args, nsWindow *target, BView *view)
{
	mTarget = target;
	mView		= view;
	printf("> %p", mTarget);
	BMessage msg;
	status_t state = msg.Unflatten((const char*)args);
//	printf("Unflatten(%lx)\n ", state);
//	hexdump((const char*)(args+15), 16);

	switch (msg.FindInt32("be:opcode")) {
	case B_INPUT_METHOD_CHANGED: {
		printf(" nsWindow::B_INPUT_METHOD_CHANGED\n");
		const char* src = msg.FindString("be:string");
		int32 srcLen = strlen(src);
		int32 dstLen = utf8_str_len(src, srcLen);
		if (mCount < dstLen) {
			delete[] mText;
			mCount = (dstLen+7);
			mText  = new PRUnichar[mCount+1];
		}
		cutf16(src, &srcLen, mText, &dstLen);
 
    if (msg.FindBool("be:confirmed")) {	// confirmed
    	if (mStateIME != NS_COMPOSITION_END)		// confirmed
   			DispatchText(mText, 0, NULL);
   		else printf("=== skip no connect\n");
   	}
   	else {
   		nsTextRange txtRuns[2];
   		PRUint32 txtCount = 2;
 	    //
 	    int32 select[2];
 	    select[0] = msg.FindInt32("be:selection", int32(0));
			select[1] = msg.FindInt32("be:selection", 1);

	 		txtRuns[0].mStartOffset = (select[0] == select[1]) ? 0 : utf8_str_len(src, select[1]);
	 		txtRuns[0].mEndOffset		= dstLen;
	    txtRuns[0].mRangeType		= NS_TEXTRANGE_CONVERTEDTEXT;
	    printf("IMEComposeText(%d %d", txtRuns[0].mStartOffset, txtRuns[0].mEndOffset);
	    if (select[0] == select[1])
	    	txtCount = 1;
	    else {
	 			txtRuns[1].mStartOffset = utf8_str_len(src, select[0]);
	 			txtRuns[1].mEndOffset		= utf8_str_len(src, select[1]);
	    	txtRuns[1].mRangeType		= NS_TEXTRANGE_SELECTEDCONVERTEDTEXT;
	    	printf(", %d %d", txtRuns[1].mStartOffset, txtRuns[1].mEndOffset);
	  	}
	    printf(")\n");
   		DispatchText(mText, txtCount, txtRuns);
	  }
	}	break;
	case B_INPUT_METHOD_LOCATION_REQUEST:
		printf(" nsWindow::B_INPUT_METHOD_LOCATION_REQUEST\n");
		if (mView && mView->LockLooper()) {
			BPoint caret(mCaret);
			BMessage reply(B_INPUT_METHOD_EVENT);
  		reply.AddInt32("be:opcode", B_INPUT_METHOD_LOCATION_REQUEST);
  		for (int32 s= 0; mText[s]; s++) { 
				reply.AddPoint("be:location_reply", mView->ConvertToScreen(caret));
				reply.AddFloat("be:height_reply", 20);
				caret.x += 14;
			}
			mIME.SendMessage(&reply);
			mView->UnlockLooper();
		}
		break;
	case B_INPUT_METHOD_STARTED:
		printf(" nsWindow::B_INPUT_METHOD_STARTED\n");
    DispatchIME(NS_COMPOSITION_START);
		DispatchIME(NS_COMPOSITION_QUERY);

		msg.FindMessenger("be:reply_to", &mIME);
		break;
	case B_INPUT_METHOD_STOPPED:
		printf(" nsWindow::B_INPUT_METHOD_STOPPED\n");
    DispatchIME(NS_COMPOSITION_END);
		delete[] mText;
		mText  = NULL;
		mCount = 0;
		break;
	default:
		printf(" RunIME(%lx) ?\n", msg.what);
	};
}

void
BezillaIME::DispatchText(PRUnichar* text, PRUint32 txtCount, nsTextRange* txtRuns)
{
	int16 *t16 = (int16*)text;
	int c = 0; while (t16[c]) c++;
	printf("DispatchText(%d %d) %p\n", c, txtCount, mTarget);
	if (mStateIME == NS_COMPOSITION_END)
		DispatchIME(NS_COMPOSITION_START);

	nsTextEvent textEvent;

	textEvent.time 		= 0;
	textEvent.isShift = 
	textEvent.isControl =
	textEvent.isAlt 	= 
	textEvent.isMeta 	= PR_FALSE;
  
	textEvent.message = textEvent.eventStructType = NS_TEXT_EVENT;
	textEvent.widget 	= NS_STATIC_CAST(nsIWidget *, mTarget);
	textEvent.point.x = 
	textEvent.point.y = 0;

	textEvent.theText 	 = text;
//textEvent.isChar		 = PR_TRUE;
	textEvent.rangeCount = txtCount;
	textEvent.rangeArray = txtRuns;

	DispatchWindowEvent(&textEvent);
}

void
BezillaIME::DispatchCancelIME()
{
		if (mText && mStateIME != NS_COMPOSITION_END) {
				printf(" nsWindow::Post B_INPUT_METHOD_STOPPED\n");
				BMessage reply(B_INPUT_METHOD_EVENT);
  			reply.AddInt32("be:opcode", B_INPUT_METHOD_STOPPED);
			if (mIME.IsValid())
				printf("mIME true\n");
			else printf("mIME false\n");
//				printf(" nsWindow::Post Key_Down\n");
//				BMessage reply(B_KEY_DOWN);
//  			reply.AddInt64("when", 0);
//  			reply.AddInt32("modifiers", 0);
//  			reply.AddInt8("byte", B_ENTER);
//  			reply.AddInt32("key", 0x47);
//  			reply.AddInt32("raw_char", B_ENTER);
//  			reply.PrintToStream();
 				status_t state = mIME.SendMessage(&reply);
 				printf("SendMessage(%lx)\n", state);

   		DispatchText(mText, 0, NULL);	// とりあえず確定する
			DispatchIME(NS_COMPOSITION_END);
		}
}

void
BezillaIME::DispatchIME(PRUint32 what)
{
	if (what == NS_COMPOSITION_END &&
	mStateIME == NS_COMPOSITION_END) {
		printf(" skip NS_COMPOSITION_END\n");
		return;
	}
    	
   nsCompositionEvent compEvent;

   compEvent.widget  = NS_STATIC_CAST(nsIWidget *, mTarget);
   compEvent.point.x =
   compEvent.point.y = 0;
   compEvent.time 	 = 0;
   compEvent.message = 
   compEvent.eventStructType = 
   compEvent.compositionMessage = what;

//   compEvent.theReply.mCursorPosition.x = -1;
//   compEvent.theReply.mCursorPosition.y = -1;

   DispatchWindowEvent(&compEvent);
   mStateIME = what;
   if (what == NS_COMPOSITION_START)
   	printf(" NS_COMPOSITION_START\n");
   else if (what == NS_COMPOSITION_END)
   	printf(" NS_COMPOSITION_END\n");
   else if (what == NS_COMPOSITION_QUERY)
   	printf(" NS_COMPOSITION_QUERY\n");
   else
   	printf(" NS_COMPOSITION_WHAT(%d)\n", what);

	if (what == NS_COMPOSITION_QUERY)
		mCaret.Set(compEvent.theReply.mCursorPosition.x,
							 compEvent.theReply.mCursorPosition.y);
}

PRBool
BezillaIME::DispatchWindowEvent(nsGUIEvent* event)
{
	nsEventStatus status;
	mTarget->DispatchEvent(event, status);
	return PR_FALSE;
}


